SystemOrganization addCategory: #'Cache-Scripting-Experiments'!

!RTPDFExporter methodsFor: '*Cache-Scripting-Experiments' stamp: 'NickPapoulias 10/27/2015 21:09'!
exportCanvas: aCanvas onFileName: aFileName
	aCanvas accept: dimensionVisitor.
	dimensionVisitor computeDimensions.
	aCanvas shapes do:[:e | e translateBy: dimensionVisitor translation].
	aCanvas accept: pdfVisitor.	
	aCanvas shapes do:[:e | e translateBy: dimensionVisitor translation * -1].
	document format: (PDFCustomFormat new size: (RTUnitConversion ptToMm: dimensionVisitor width) point @ (RTUnitConversion ptToMm: dimensionVisitor height) point).
	self exportFilename: aFileName.
	! !

!RTPDFExporter methodsFor: '*Cache-Scripting-Experiments' stamp: 'NickPapoulias 10/27/2015 21:06'!
exportFilename: aFileName 

	| fileRef fileStream |
	
	fileRef :=  FileReference / aFileName.
	fileStream := fileRef writeStream.
	[ self exportOnStream: fileStream. ] 
		ensure: [ fileStream notNil ifTrue: [ fileStream close ] ]! !

TestCase subclass: #CachingVariableSanityTest
	instanceVariableNames: 'collection'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Cache-Scripting-Experiments'!

!CachingVariableSanityTest methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/24/2015 17:59'!
testEquality

	self assert: (TempCachingVariable new 
							class: Object 
							name: 'someCache' 
							method: '#Object>>someMethod' ) = (TempCachingVariable new 
																			class: Object 
																			name: 'someCache' 
																			method: '#Object>>someMethod' ).
	self deny: (TempCachingVariable new 
							class: Object 
							name: 'someCache' 
							method: '#Object>>someMethod' ) = (TempCachingVariable new 
																			class: Object 
																			name: 'someCache' 
																			method: '#Object>>someMethod2' )
	! !

!CachingVariableSanityTest methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/24/2015 18:12'!
testImportExport

	collection := OrderedCollection new.

	1 to: 10 do: [: i | collection add: (TempCachingVariable new 
												class: Object 
												name: 'someCache' 
												method: '#Object>>someMethod' ) ].
											
	FLSerializer serialize: collection toFileNamed: self class name" , '-' , DateAndTime now asString" , '.FUEL'.
	collection := FLMaterializer materializeFromFileNamed: self class name" , '-' , DateAndTime now asString" , '.FUEL'.
	self assert: collection size = 10.
	self assert: collection asSet size = 1.									
	! !

!CachingVariableSanityTest methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/24/2015 18:01'!
testSet

	collection := OrderedCollection new.

	1 to: 10 do: [: i | collection add: (TempCachingVariable new 
												class: Object 
												name: 'someCache' 
												method: '#Object>>someMethod' ) ].
	self assert: collection asSet size = 1.! !

Object subclass: #CacheCollection
	instanceVariableNames: 'imageInfo collection'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Cache-Scripting-Experiments'!

CacheCollection subclass: #AnalysisInformallyAnotatedCaches
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Cache-Scripting-Experiments'!

AnalysisInformallyAnotatedCaches subclass: #AnalysisInformallyAnotatedBuffers
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Cache-Scripting-Experiments'!

!AnalysisInformallyAnotatedBuffers methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/22/2015 13:08'!
isInformallyAnnotated: iv

	^ (iv includesSubstring: 'buf' caseSensitive: false) ! !

!AnalysisInformallyAnotatedCaches methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/22/2015 12:58'!
cachesForClass: aClass

	^ aClass instanceVariables ifNil: [ {  } ] ifNotNilDo: 
		[:ivs | (ivs select: [ :iv | self isInformallyAnnotated: iv]) asOrderedCollection ]
! !

!AnalysisInformallyAnotatedCaches methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/22/2015 13:00'!
cachesForClassInClassVars: aClass

	^ aClass classVarNames ifNil: [ {  } ] ifNotNilDo: 
		[:ivs | (ivs select: [ :iv | self isInformallyAnnotated: iv ]) asOrderedCollection ]
! !

!AnalysisInformallyAnotatedCaches methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/22/2015 13:00'!
cachesForMethod: aMethod

	^ aMethod tempNames ifNil: [ {  } ] ifNotNilDo: 
		[:temps | (temps select: [ :temp | self isInformallyAnnotated: temp]) asOrderedCollection ]! !

!AnalysisInformallyAnotatedCaches methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/24/2015 17:54'!
checkForClassVarsIn: class

	class isMeta ifFalse: [
				(self cachesForClassInClassVars: class)
					ifNotEmptyDo: [:caches | 
							caches do: [ :cache | collection add: (SharedCachingVariable new class: class  name: cache ) ]]
			]! !

!AnalysisInformallyAnotatedCaches methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/24/2015 17:54'!
checkForIVsIn: class

	(self cachesForClass: class) ifNotEmptyDo: [:caches | 
		caches do: [ :cache |
				|cachingVar| 
				class isMeta
					ifTrue: [ cachingVar := ClassSideCachingVariable new class: class name: cache  ]
					ifFalse: [ cachingVar := InstanceSideCachingVariable new class: class name: cache  ]. 
					collection add: cachingVar ]].! !

!AnalysisInformallyAnotatedCaches methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/24/2015 17:54'!
checkForTempsIn: class

	class methods do: [ :method |
				(self cachesForMethod: method) 
					ifNotEmptyDo: [:caches | 
						caches do: [ :cache | collection add: (TempCachingVariable new class: method methodClass name: cache method: method ) ]]
				]! !

!AnalysisInformallyAnotatedCaches methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/20/2015 17:37'!
gatherSet

	ProtoObject allSubclasses do: [ :class |
		class package ~= self class package ifTrue: [   
			self checkForIVsIn: class. "both instance side and class side for metaclasses"
			self checkForClassVarsIn: class.
			self checkForTempsIn: class.
		]
	].
	! !

!AnalysisInformallyAnotatedCaches methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/22/2015 13:04'!
isInformallyAnnotated: iv

	^ (iv includesSubstring: 'cach' caseSensitive: false) 
		"[ (iv includesSubstring: 'buf' caseSensitive: false) ]"! !

!CacheCollection class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/23/2015 14:47'!
importFromFuel: aFileName

	^ FLMaterializer materializeFromFileNamed: aFileName.
	! !

!CacheCollection methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/24/2015 18:07'!
collection

	^ collection! !

!CacheCollection methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/24/2015 17:55'!
collectionAsSet

	^ collection asSet "has already removed duplicates -- and avoids problems when doing collect: and counting results"! !

!CacheCollection methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/24/2015 19:01'!
exportToFuel

	FLSerializer serialize: self toFileNamed: self class name" , '-' , DateAndTime now asString" , '.FUEL'.! !

!CacheCollection methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/23/2015 11:11'!
gatherProcessAndExport

	self gatherSet postProcessing exportToFuel.! !

!CacheCollection methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/20/2015 18:02'!
gatherSet

	self subclassResponsibility ! !

!CacheCollection methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/24/2015 18:10'!
hasDuplicates

	^ (collection size = self collection asSet size) not! !

!CacheCollection methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/23/2015 11:22'!
imageInfo

	^ imageInfo ! !

!CacheCollection methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/24/2015 17:56'!
initialize

	super initialize.
	collection := OrderedCollection new.
	imageInfo := ImageInfo new.! !

!CacheCollection methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/24/2015 17:54'!
postProcessing

	collection do: [ :cachingVariable | cachingVariable postProcessing ]! !

CacheCollection subclass: #CaseStudySystemWide
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Cache-Scripting-Experiments'!

CacheCollection subclass: #ValidationGlobalStateCaches
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Cache-Scripting-Experiments'!

Object subclass: #CachingVarFigures
	instanceVariableNames: 'cacheCollection imageInfo container cacheSet'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Cache-Scripting-Experiments'!

!CachingVarFigures methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/26/2015 20:37'!
cachingImplPropertiesGraph! !

!CachingVarFigures methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/26/2015 20:39'!
cachingProtocolsGraph! !

!CachingVarFigures methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/26/2015 20:38'!
cachingSelectorsGraph! !

!CachingVarFigures methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/26/2015 20:37'!
cachingValuesGraph! !

!CachingVarFigures methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/26/2015 20:36'!
cachingVarsDensityStats! !

!CachingVarFigures methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/27/2015 21:08'!
cachingVarsKinds

	| b  stats  |
	b := RTDoubleBarBuilder new.
	stats := Dictionary new.
	stats 
		at: 'Instance Variables' 
		put: (cacheSet select: [:cvar | cvar class = InstanceSideCachingVariable]) size. 
	stats 
		at: 'Class-Side Variables' 
		put: (cacheSet select: [:cvar | cvar class = ClassSideCachingVariable]) size.
	stats 
		at: 'Shared Variables' 
		put: (cacheSet select: [:cvar | cvar class = SharedCachingVariable]) size.			
	stats 
		at: 'Temporary Variables' 
		put: (cacheSet select: [:cvar | cvar class = TempCachingVariable]) size.
	b points: stats keys.
	b bottomValue: [:x | stats at: x] titled: 'occurences'.
	"b topValue: #numberOfVariables".
	"b axisX noTick; noLabel."
	b build.
	b view setDefaultMenu; open.
	RTPDFExporter new exportCanvas: b view canvas onFileName: '/home/parseval/Desktop/SANER/cachingVarsKinds.pdf'.
	^ b view! !

!CachingVarFigures methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/26/2015 20:34'!
cachingVarsPackageGroupStats! !

!CachingVarFigures methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/26/2015 20:31'!
cachingVarsPerPackageGroupPie! !

!CachingVarFigures methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/26/2015 20:32'!
cachingVarsToNumberOfPackageGroups! !

!CachingVarFigures methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/27/2015 20:32'!
importFromFuel: aFileName for: aCacheCollectionSubclass

	container := (aCacheCollectionSubclass importFromFuel: aFileName).
	cacheCollection := container collection.
	cacheSet := container collectionAsSet.
	imageInfo := container imageInfo. 
	! !

Object subclass: #CachingVariable
	instanceVariableNames: 'hash packageGroup package class name methods testClass values selectors protocols exposable patterns'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Cache-Scripting-Experiments'!

!CachingVariable commentStamp: 'NickPapoulias 10/23/2015 12:51' prior: 0!
Please comment me using the following template inspired by Class Responsibility Collaborator (CRC) design:

For the Class part:  State a one line summary. For example, "I represent a paragraph of text".

For the Responsibility part: Three sentences about my main responsibilities - what I do, what I know.

For the Collaborators Part: State my main collaborators and one line about how I interact with them. 

Public API and Key Messages

- message one   
- message two 
- (for bonus points) how to create instances.

   One simple example is simply gorgeous.
 
Internal Representation and Key Implementation Points.

    Instance Variables
	class: <Symbol> Class that contains the cache. Scope is determined in subclasses. 
			eg #SystemDictionary
	exposable: <Boolean> If it is returned directly, or as a msg-send with the cache directly as the receiver (heuristic) 
			eg true
	methods: <OrderedCollection> Methods that have VariableNode with the cache. 
			eg. an OrderedCollection(#'SystemDictionary>>#flushClassNameCache' #'SystemDictionary>>#nonClassNames' #'SystemDictionary>>#fillCaches') 
	name:	<Symbol> The name of the cache
			eg. #cachedNonClassNames
	package: <Symbol>  The package of the containing class
			eg.  #System-Support
	packageGroup: <Symbol> The package group of the containing class
			eg. #System
	patterns: <OrderedCollection> Todo Guille:
			eg. (probably smth like) an asOrderedCollection (#'SystemDictionary>>#flushClassNameCache' -> #flushing-pattern)) 
	protocols: <Set> The statically retrieved protocols corresponding to selectors directly sent to the cache: 
			eg.  a Set(#testing)
	selectors:	<Set> Selectors that were send directly to the cache
			eg.  a Set(#ifNil:)
	testClass:	<Symbol> The corresponding testClass if it exists
			eg.  #SystemDictionaryTest
	values:	<Set> The values directly assigned as literal or as Global new 
			eg. a Set(nil)


    Implementation Points!

!CachingVariable methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/20/2015 19:54'!
= anObject

	^ ((self class = anObject class) and: [ class = anObject cacheClass ]) and: [ name = anObject name ]! !

!CachingVariable methodsFor: 'accessing' stamp: 'NickPapoulias 10/20/2015 17:36'!
cacheClass
	^ class! !

!CachingVariable methodsFor: 'accessing' stamp: 'NickPapoulias 10/24/2015 18:06'!
class: aClass name: aName

	name := aName asSymbol. "just to be sure"
	class := aClass name asSymbol.
	package := aClass package name asSymbol.
	packageGroup := (package asString findTokens: '-') first asSymbol.

	! !

!CachingVariable methodsFor: 'accessing' stamp: 'NickPapoulias 10/22/2015 13:22'!
exposable
	^ exposable! !

!CachingVariable methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/23/2015 14:43'!
hash

	^ hash ifNil: [ hash := (self class hash bitXor: class hash) bitXor: name hash]! !

!CachingVariable methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/22/2015 12:45'!
liveClass

	^ (class includesSubstring: ' class') "metaclass"
			ifTrue: [ (Smalltalk globals at: ((class findTokens: ' ') first) asSymbol) class ]
			ifFalse: [ (Smalltalk globals at: class asSymbol) ].
		! !

!CachingVariable methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/23/2015 13:51'!
liveClassFor: aClassPrintString

	^ (aClassPrintString includesSubstring: ' class') "metaclass"
			ifTrue: [ (Smalltalk globals at: ((aClassPrintString findTokens: ' ') first) asSymbol) class ]
			ifFalse: [ (Smalltalk globals at: aClassPrintString asSymbol) ].
		! !

!CachingVariable methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/23/2015 13:51'!
liveMethods

	^ methods collect: [ :method | |tokens |
		tokens := method asString findTokens: '>>#'.
		(self liveClassFor: tokens first) >> tokens last asSymbol]! !

!CachingVariable methodsFor: 'accessing' stamp: 'NickPapoulias 10/20/2015 17:36'!
methods
	^ methods! !

!CachingVariable methodsFor: 'accessing' stamp: 'NickPapoulias 10/20/2015 17:37'!
name
	^ name! !

!CachingVariable methodsFor: 'accessing' stamp: 'NickPapoulias 10/22/2015 13:19'!
package
	^ package! !

!CachingVariable methodsFor: 'accessing' stamp: 'NickPapoulias 10/22/2015 13:19'!
packageGroup
	^ packageGroup! !

!CachingVariable methodsFor: 'accessing' stamp: 'NickPapoulias 10/22/2015 13:23'!
patterns
	^ patterns! !

!CachingVariable methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/22/2015 15:06'!
postProcessing

	"calculate values, selectors and patterns"
	self postProcessingTestClass.
	self postProcessingMethods.
	self postProcessingValues.
	self postProcessingSelectors.
	self postProcessingProtocols. "less concrete in static analysis"
	self postProcessingExposable.
	self postProcessingPatterns.! !

!CachingVariable methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/23/2015 12:45'!
postProcessingExposable

	self liveMethods do: [ :method | 
		method ast nodesDo: [ :node |  
			((node isReturn and: [ node value isVariable ]) and: [ node value name = name ])
				ifTrue: [ ^ exposable := true ].	
			(((node isReturn and: [ node value isMessage ]) and:  [ node value receiver isVariable ]) and: [ node value receiver name = name ]) "heuristic"
				ifTrue: [ ^ exposable := true ]
		]	
	].
	^ exposable := false! !

!CachingVariable methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/23/2015 13:52'!
postProcessingMethods

	"calculate values, selectors and patterns"
	methods := (self liveClass methods asOrderedCollection 
						select: [:method | 
							(method variableNodes select: [ :node | node name = name ]) isEmpty not ])
						collect: [ :method | method printString asSymbol ]
	! !

!CachingVariable methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/21/2015 18:29'!
postProcessingPatterns
! !

!CachingVariable methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/22/2015 12:52'!
postProcessingProtocols

	protocols := (selectors flatCollect: [:selector | 
		(SystemNavigation default allImplementorsOf: selector) 
			collect: [ :rgMethod | rgMethod protocol ]  ]) asSet! !

!CachingVariable methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/22/2015 12:53'!
postProcessingSelectors

	|nodes|

	nodes := ((self liveMethods flatCollect: [:method | method variableNodes ])
		select: [ :node | node name = name ])
		select: [ :node | node parent isMessage].
		
	selectors := (nodes collect: [ :node | node parent selector ]) asSet.
	
	! !

!CachingVariable methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/27/2015 13:57'!
postProcessingTestClass

	|lookFor|

	lookFor := (self liveClass theNonMetaClass asString , 'Test') asSymbol.
	(Smalltalk globals keys includes: lookFor)
		ifTrue: [ testClass := lookFor ]
		ifFalse: [  
			(self liveClass inheritsFrom: TestCase)
				ifTrue: [ testClass := self liveClass name asSymbol ]
		]! !

!CachingVariable methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/22/2015 14:00'!
postProcessingValues

	|nodes|

	nodes := ((self liveMethods flatCollect: [:method | method assignmentNodes ])
		select: [ :assignNode | assignNode variable name = name ])
		collect: [ :assignNode | assignNode value].
	values := OrderedCollection new.
	values addAll: ((nodes select: [:node | node isLiteralNode ]) collect: [ :node | node value ]).
	values addAll: 
		(((nodes select: [:node | node isMessage ]) 
			select: [ :node | (node receiver isVariable and: [ node receiver isGlobal ]) and: [node selector = #new] ])
			collect: [:node | node receiver name]).
	values := values asSet! !

!CachingVariable methodsFor: 'accessing' stamp: 'NickPapoulias 10/22/2015 13:22'!
protocols
	^ protocols! !

!CachingVariable methodsFor: 'accessing' stamp: 'NickPapoulias 10/22/2015 13:22'!
selectors
	^ selectors! !

!CachingVariable methodsFor: 'accessing' stamp: 'NickPapoulias 10/22/2015 13:19'!
testClass
	^ testClass! !

!CachingVariable methodsFor: 'accessing' stamp: 'NickPapoulias 10/22/2015 13:22'!
values
	^ values! !

CachingVariable subclass: #ClassSideCachingVariable
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Cache-Scripting-Experiments'!

CachingVariable subclass: #InstanceSideCachingVariable
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Cache-Scripting-Experiments'!

CachingVariable subclass: #SharedCachingVariable
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Cache-Scripting-Experiments'!

!SharedCachingVariable methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/23/2015 13:48'!
postProcessingMethods

	super postProcessingMethods.
	methods ifNil: [ methods := OrderedCollection new ].
	methods addAll: ((self liveClass class methods "class side"
						select: [:method | 
							(method variableNodes select: [ :node | node name = name ]) isEmpty not ])
						collect: [ :method | method printString asSymbol ]).! !

CachingVariable subclass: #TempCachingVariable
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Cache-Scripting-Experiments'!

!TempCachingVariable methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/20/2015 19:55'!
= anObject

	^ (super = anObject) and: [ methods = anObject methods ]! !

!TempCachingVariable methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/24/2015 18:06'!
class: aClass name: aName method: aMethod

	super class: aClass name: aName.
	methods := { aMethod printString asSymbol } asOrderedCollection.! !

!TempCachingVariable methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/23/2015 14:42'!
hash
	
	^ hash ifNil: [ hash := super hash bitXor: methods first hash ] "some spookey bug with fuel here check"! !

!TempCachingVariable methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/21/2015 15:58'!
postProcessingMethods
	
	"nothing to do only one method accesses me"! !

Object subclass: #ImageInfo
	instanceVariableNames: 'version packagesToNumberOfClasses packageGroupsToNumberOfClasses'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Cache-Scripting-Experiments'!

!ImageInfo methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/23/2015 11:39'!
initialize

	version := SystemVersion current asString.
	packagesToNumberOfClasses := (ProtoObject allSubclasses 
		collect: [ :class | class package name asSymbol -> class package classes size ]) asSet asDictionary.
	packagesToNumberOfClasses removeKey: self class package name asSymbol. "don't count the experiment package"
	packageGroupsToNumberOfClasses := Dictionary new.
	packagesToNumberOfClasses keysAndValuesDo:
		 [ :package :numberOfClasses | 
			|before packageGroup|
			packageGroup := (package asString findTokens: '-') first asSymbol.
			before := (packageGroupsToNumberOfClasses 
				at: packageGroup
				ifAbsentPut: 0).
			packageGroupsToNumberOfClasses 
				at: packageGroup
			  	put: before + numberOfClasses.
		].! !
